﻿/**
 * ...
 * @author Georg Graf
 */

package com.georg;
import com.georg.tools.ComplexNumber;
import com.georg.tools.RGB;
import flash.display.Bitmap;
import flash.display.BitmapData;
import flash.display.Sprite;
import flash.events.Event;
import flash.events.TimerEvent;
import flash.geom.ColorTransform;
import flash.utils.Timer;

class Mandelbrot extends Bitmap
{
	// draw preferences
	public var colorTable(default, null):Array<Int>;
	public var colorAngle(default, null):Int;
	public var iterations(default, null):Int;
	public var scale(default, null):Float;
	public var offsetX:Float;
	public var offsetY:Float;
	
	// drawIncremental variables
	private var timer:Timer;
	private var tileSteps:Int;
	private var tileWidth:Int;
	private var tileHeight:Int;
	private var tileStartX:Int;
	private var tileStartY:Int;
	private var onCompleteFunction:Void->Void;
	
	// drawPixelLike variables
	private var pixSteps:Int;
	private var pixTimer:Timer;
	private var pixWidth:Int;
	private var pixHeight:Int;
	private var pixStartX:Int;
	private var pixStartY:Int;
	
	// draw variables
	private var cnZ:ComplexNumber;
	private var count:Int;
	
	public function new(width:Int, height:Int) 
	{
		offsetX = 0;
		offsetY = 0;
		
		super(new BitmapData(width, height, false, 0));
	}
	
	inline public function drawIncremental(steps:Int = 4, delay:Float = 10, iterations:Int = 20, scale:Float = .01, colorAngle:Int = 10, onComplete:Void->Void = null):Void
	{
		this.iterations = iterations;
		this.scale = scale;
		this.colorAngle = colorAngle;
		this.onCompleteFunction = onComplete;
		
		if (timer != null)
		{
			timer.stop();
			timer.removeEventListener(TimerEvent.TIMER, drawPiece);
			timer = null;
		}
		
		createColorTable();
		
		// set calculation variables
		tileSteps = steps;
		tileWidth = cast(width / tileSteps);
		tileHeight = cast(height / tileSteps);
		
		// start timing
		timer = new Timer(delay, steps * steps);
		timer.addEventListener(TimerEvent.TIMER, drawPiece);
		timer.start();
	}
	
	inline public function drawPixelIncremental(steps:Int = 10, delay:Float = 10, iterations:Int = 20, scale:Float = .01, colorAngle:Int = 10, onComplete:Void->Void = null):Void
	{
		this.iterations = iterations;
		this.scale = scale;
		this.colorAngle = colorAngle;
		this.onCompleteFunction = onComplete;
		
		if (pixTimer != null)
		{
			pixTimer.stop();
			pixTimer.removeEventListener(TimerEvent.TIMER, drawPiece);
			pixTimer = null;
		}
		
		createColorTable();
		
		// set calculation variables
		pixSteps = steps;
		pixWidth = cast(width / pixSteps);
		pixHeight = cast(height / pixSteps);
		
		// start timing
		pixTimer = new Timer(delay, pixWidth * pixHeight);
		pixTimer.addEventListener(TimerEvent.TIMER, drawPixelLike);
		pixTimer.start();
	}
	
	inline public function draw(iterations:Int = 20, scale:Float = .01, colorAngle:Int = 10):Void
	{
		this.iterations = iterations;
		this.scale = scale;
		this.colorAngle = colorAngle;
		
		createColorTable();
		drawSegment(cast( -(width / 2), Int), 	// -half width
					cast((width / 2), Int), 	// +half width
					cast( -(height / 2), Int), 	// -half height
					cast((height / 2), Int));	// +half height
	}
	
	public function stop():Void
	{
		if (timer != null)		timer.stop();
		if (pixTimer != null)	pixTimer.stop();
	}
	
	public function drag(offX:Float, offY:Float):Void
	{
		offsetX -= offX;
		offsetY -= offY;
		
		bitmapData.scroll(cast(offX, Int), cast(offY, Int));
	}

	private function drawPiece(e:TimerEvent):Void
	{
		tileStartX = cast( (((timer.currentCount - 1) % tileSteps) * tileWidth) - (width / 2), Int);
		tileStartY = cast( (Math.floor((timer.currentCount - 1) / tileSteps) * tileHeight) - (height / 2), Int);
		
		drawSegment(tileStartX, 
					tileStartX + tileWidth,
					tileStartY,
					tileStartY + tileHeight);
					
		if (timer.currentCount == timer.repeatCount)
			onCompleteFunction();
	}
	
	private function drawPixelLike(e:TimerEvent):Void
	{
		pixStartX = cast( (((pixTimer.currentCount - 1) % pixWidth)) - (width / 2), Int);
		pixStartY = cast( (Math.floor((pixTimer.currentCount - 1) / pixWidth)) - (height / 2), Int);
		
		for (iX in 0...pixSteps)
			for (iY in 0...pixSteps)
				drawPixel(pixStartX + (iX * pixWidth), pixStartY + (iY * pixHeight));
	}
	
	private function createColorTable():Void
	{
		colorTable = new Array<Int>();
				
		for (i in 0...iterations)
			colorTable.push(RGB.hsv(colorAngle, 1, (i / iterations)));
	}
	
	/** the core-function to draw a segment of the Mandelbrot **/
	inline private function drawSegment(startX:Int, endX:Int, startY:Int, endY:Int):Void
	{
		// create image 
		for (iX in startX...endX)
			for (iY in startY...endY)
				drawPixel(iX, iY);
	}
	
	/** the core-function to draw one pixel of the Mandelbrot **/
	inline private function drawPixel(posX:Int, posY:Int):Void
	{
		cnZ = new ComplexNumber(0, 0);
		count = 0;
		
		while(count < iterations)
		{
			// z(n+1) = z(n)^2 + c
			cnZ = (cnZ.power()).add(new ComplexNumber((posX + offsetX) * scale, (posY + offsetY) * scale));
			count++;
			
			if (cnZ.abs() > 2)
				break;
		}
		bitmapData.setPixel(posX + cast(width / 2, Int), posY + cast(height / 2, Int), colorTable[count]);
	}
	
	/** completely removes all dependencies an stops working **/
	public function destroy():Void
	{
		if (timer != null)		
		{
			pixTimer.stop();
			pixTimer.removeEventListener(TimerEvent.TIMER, drawPiece);
			pixTimer = null;
		}
		if (pixTimer != null)	
		{
			pixTimer.stop();
			pixTimer.removeEventListener(TimerEvent.TIMER, drawPiece);
			pixTimer = null;
		}
	}
}